@@ -39,6 +39,11 @@
 #include <gio/gdesktopappinfo.h>
 #include <gio/gunixfdlist.h>
 
+#if 0
+#include <libupower-glib/upower.h>
+#endif
+#include <gdesktop-enums.h>
+
 #ifdef HAVE_GUDEV
 #include <gudev/gudev.h>
 #endif
@@ -75,6 +80,7 @@
 #define CUSTOM_BINDING_SCHEMA SETTINGS_BINDING_DIR ".custom-keybinding"
 
 #define SHELL_GRABBER_RETRY_INTERVAL 1
+#define OSD_ALL_OUTPUTS -1
 
 static const gchar introspection_xml[] =
 "<node name='/org/gnome/SettingsDaemon/MediaKeys'>"
@@ -95,9 +101,10 @@ static const gchar introspection_xml[] =
 "</node>";
 
 #define SETTINGS_INTERFACE_DIR "org.gnome.desktop.interface"
+#define SETTINGS_POWER_DIR "org.gnome.settings-daemon.plugins.power"
 #define SETTINGS_XSETTINGS_DIR "org.gnome.settings-daemon.plugins.xsettings"
-#define SETTINGS_TOUCHPAD_DIR "org.gnome.settings-daemon.peripherals.touchpad"
-#define TOUCHPAD_ENABLED_KEY "touchpad-enabled"
+#define SETTINGS_TOUCHPAD_DIR "org.gnome.desktop.peripherals.touchpad"
+#define TOUCHPAD_ENABLED_KEY "send-events"
 #define HIGH_CONTRAST "HighContrast"
 
 #define VOLUME_STEP 6           /* percents for one volume button press */
@@ -118,7 +125,7 @@ typedef struct {
 
 typedef struct {
         MediaKeyType key_type;
-        ShellKeyBindingMode modes;
+        ShellActionMode modes;
         const char *settings_key;
         const char *hard_coded;
         char *custom_path;
@@ -154,6 +161,16 @@ struct GsdMediaKeysManagerPrivate
         char            *icon_theme;
         char            *gtk_theme;
 
+#if 0
+        /* Power stuff */
+        GSettings       *power_settings;
+        GDBusProxy      *power_proxy;
+        GDBusProxy      *power_screen_proxy;
+        GDBusProxy      *power_keyboard_proxy;
+        UpDevice        *composite_device;
+        char            *chassis_type;
+#endif
+
         /* Shell stuff */
         GsdShell        *shell_proxy;
         ShellKeyGrabber *key_grabber;
@@ -297,13 +314,14 @@ static void
 show_osd (GsdMediaKeysManager *manager,
           const char          *icon,
           const char          *label,
-          int                  level)
+          int                  level,
+          int                  output_id)
 {
         if (manager->priv->shell_proxy == NULL)
                 return;
 
         shell_show_osd (manager->priv->shell_proxy,
-                        icon, label, level, -1);
+                        icon, label, level, output_id);
 }
 
 static const char *
@@ -552,7 +570,7 @@ media_key_new_for_path (GsdMediaKeysManager *manager,
 
         key = g_new0 (MediaKey, 1);
         key->key_type = CUSTOM_KEY;
-        key->modes = GSD_KEYBINDING_MODE_LAUNCHER;
+        key->modes = GSD_ACTION_MODE_LAUNCHER;
         key->custom_path = g_strdup (path);
         key->custom_command = command;
 
@@ -837,7 +855,7 @@ do_eject_action (GsdMediaKeysManager *manager)
         }
 
         /* Show OSD */
-        show_osd (manager, "media-eject-symbolic", NULL, -1);
+        show_osd (manager, "media-eject-symbolic", NULL, -1, OSD_ALL_OUTPUTS);
 
         /* Clean up the drive selection and exit if no suitable
          * drives are found */
@@ -912,7 +930,7 @@ static void
 do_touchpad_osd_action (GsdMediaKeysManager *manager, gboolean state)
 {
         show_osd (manager, state ? "input-touchpad-symbolic"
-                                 : "touchpad-disabled-symbolic", NULL, -1);
+                                 : "touchpad-disabled-symbolic", NULL, -1, OSD_ALL_OUTPUTS);
 }
 
 static void
@@ -927,11 +945,15 @@ do_touchpad_action (GsdMediaKeysManager *manager)
         }
 
         settings = g_settings_new (SETTINGS_TOUCHPAD_DIR);
-        state = g_settings_get_boolean (settings, TOUCHPAD_ENABLED_KEY);
+        state = (g_settings_get_enum (settings, TOUCHPAD_ENABLED_KEY) ==
+                 G_DESKTOP_DEVICE_SEND_EVENTS_ENABLED);
 
         do_touchpad_osd_action (manager, !state);
 
-        g_settings_set_boolean (settings, TOUCHPAD_ENABLED_KEY, !state);
+        g_settings_set_enum (settings, TOUCHPAD_ENABLED_KEY,
+                             !state ?
+                             G_DESKTOP_DEVICE_SEND_EVENTS_ENABLED :
+                             G_DESKTOP_DEVICE_SEND_EVENTS_DISABLED);
         g_object_unref (settings);
 }
 
@@ -1028,9 +1050,9 @@ update_dialog (GsdMediaKeysManager *manager,
              g_strcmp0 (port->port, "analog-output") != 0)) {
                 device = gvc_mixer_control_lookup_device_from_stream (manager->priv->volume, stream);
                 show_osd (manager, icon,
-                          gvc_mixer_ui_device_get_description (device), vol);
+                          gvc_mixer_ui_device_get_description (device), vol, OSD_ALL_OUTPUTS);
         } else {
-                show_osd (manager, icon, NULL, vol);
+                show_osd (manager, icon, NULL, vol, OSD_ALL_OUTPUTS);
         }
 
         if (quiet == FALSE && sound_changed != FALSE && muted == FALSE) {
@@ -1477,7 +1499,7 @@ gsd_media_player_key_pressed (GsdMediaKeysManager *manager,
         if (!have_listeners) {
                 if (!mpris_controller_key (manager->priv->mpris_controller, key)) {
                         /* Popup a dialog with an (/) icon */
-                        show_osd (manager, "action-unavailable-symbolic", NULL, -1);
+                        show_osd (manager, "action-unavailable-symbolic", NULL, -1, OSD_ALL_OUTPUTS);
                 }
 		return TRUE;
         }
@@ -1656,7 +1678,7 @@ do_video_rotate_lock_action (GsdMediaKeysManager *manager,
         g_object_unref (settings);
 
         show_osd (manager, locked ? "rotation-locked-symbolic"
-                                  : "rotation-allowed-symbolic", NULL, -1);
+                                  : "rotation-allowed-symbolic", NULL, -1, OSD_ALL_OUTPUTS);
 }
 
 static void
@@ -1786,13 +1808,9 @@ power_action (GsdMediaKeysManager *manager,
 
 static void
 do_config_power_action (GsdMediaKeysManager *manager,
-                        const gchar *config_key,
-                        gboolean in_lock_screen)
+                        GsdPowerActionType   action_type,
+                        gboolean             in_lock_screen)
 {
-        GsdPowerActionType action_type;
-
-        action_type = g_settings_get_enum (manager->priv->power_settings,
-                                           config_key);
         switch (action_type) {
         case GSD_POWER_ACTION_SUSPEND:
                 power_action (manager, "Suspend", !in_lock_screen);
@@ -1816,28 +1834,59 @@ do_config_power_action (GsdMediaKeysManager *manager,
         }
 }
 
+static void
+do_config_power_button_action (GsdMediaKeysManager *manager,
+                               gboolean             in_lock_screen)
+{
+        GsdPowerButtonActionType action_type;
+
+        /* Always power off VMs when power off is pressed in the menus */
+        if (g_strcmp0 (manager->priv->chassis_type, "vm") == 0) {
+                power_action (manager, "PowerOff", !in_lock_screen);
+                return;
+        }
+
+        /* Always suspend tablets */
+        if (g_strcmp0 (manager->priv->chassis_type, "tablet") == 0) {
+                power_action (manager, "Suspend", !in_lock_screen);
+                return;
+        }
+
+        action_type = g_settings_get_enum (manager->priv->power_settings, "power-button-action");
+        switch (action_type) {
+        case GSD_POWER_BUTTON_ACTION_SUSPEND:
+                do_config_power_action (manager, GSD_POWER_ACTION_SUSPEND, in_lock_screen);
+                break;
+        case GSD_POWER_BUTTON_ACTION_HIBERNATE:
+                do_config_power_action (manager, GSD_POWER_ACTION_HIBERNATE, in_lock_screen);
+                break;
+        case GSD_POWER_BUTTON_ACTION_NOTHING:
+                /* do nothing */
+                break;
+        }
+}
+
 static void
 update_brightness_cb (GObject             *source_object,
                       GAsyncResult        *res,
                       gpointer             user_data)
 {
         GError *error = NULL;
-        int percentage;
-        GVariant *new_percentage;
+        int percentage, output_id;
+        GVariant *variant;
         GsdMediaKeysManager *manager = GSD_MEDIA_KEYS_MANAGER (user_data);
         const char *icon, *debug;
 
+        /* update the dialog with the new value */
         if (G_DBUS_PROXY (source_object) == manager->priv->power_keyboard_proxy) {
-                icon = "keyboard-brightness-symbolic";
                 debug = "keyboard";
         } else {
-                icon = "display-brightness-symbolic";
                 debug = "screen";
         }
 
-        new_percentage = g_dbus_proxy_call_finish (G_DBUS_PROXY (source_object),
-                                                   res, &error);
-        if (new_percentage == NULL) {
+        variant = g_dbus_proxy_call_finish (G_DBUS_PROXY (source_object),
+                                        res, &error);
+        if (variant == NULL) {
                 g_warning ("Failed to set new %s percentage: %s",
                            debug, error->message);
                 g_error_free (error);
@@ -1845,9 +1894,17 @@ update_brightness_cb (GObject             *source_object,
         }
 
         /* update the dialog with the new value */
-        g_variant_get (new_percentage, "(i)", &percentage);
-        show_osd (manager, icon, NULL, percentage);
-        g_variant_unref (new_percentage);
+        if (G_DBUS_PROXY (source_object) == manager->priv->power_keyboard_proxy) {
+                icon = "keyboard-brightness-symbolic";
+                output_id = -1;
+                g_variant_get (variant, "(i)", &percentage);
+        } else {
+                icon = "display-brightness-symbolic";
+                g_variant_get (variant, "(ii)", &percentage, &output_id);
+        }
+
+        show_osd (manager, icon, NULL, percentage, output_id);
+        g_variant_unref (variant);
 }
 
 static void
@@ -1921,11 +1978,28 @@ do_battery_action (GsdMediaKeysManager *manager)
 
         if (kind == UP_DEVICE_KIND_UPS || kind == UP_DEVICE_KIND_BATTERY) {
                 g_debug ("showing battery level OSD");
-                show_osd (manager, icon_name, NULL, percentage);
+                show_osd (manager, icon_name, NULL, percentage, OSD_ALL_OUTPUTS);
         }
 
         g_free (icon_name);
 }
+#else
+static void
+do_config_power_button_action (GsdMediaKeysManager *manager,
+                               gboolean             in_lock_screen)
+{}
+static void
+do_config_power_action (GsdMediaKeysManager *manager,
+                        GsdPowerActionType   action_type,
+                        gboolean             in_lock_screen)
+{}
+static void
+do_brightness_action (GsdMediaKeysManager *manager,
+                      MediaKeyType type)
+{}
+static void
+do_battery_action (GsdMediaKeysManager *manager)
+{}
 #endif
 
 static void
@@ -2006,11 +2080,14 @@ do_custom_action (GsdMediaKeysManager *manager,
 static gboolean
 do_action (GsdMediaKeysManager *manager,
            guint                deviceid,
+           guint                mode,
            MediaKeyType         type,
            gint64               timestamp)
 {
         g_debug ("Launching action for key type '%d' (on device id %d)", type, deviceid);
 
+        gboolean power_action_noninteractive = (POWER_KEYS_MODE_NO_DIALOG & mode);
+
         switch (type) {
         case TOUCHPAD_KEY:
                 do_touchpad_action (manager);
@@ -2130,19 +2207,26 @@ do_action (GsdMediaKeysManager *manager,
 		do_toggle_contrast_action (manager);
 		break;
         case POWER_KEY:
+                do_config_power_button_action (manager, power_action_noninteractive);
+                break;
         case SLEEP_KEY:
+                do_config_power_action (manager, GSD_POWER_ACTION_HIBERNATE, power_action_noninteractive);
+                break;
         case SUSPEND_KEY:
+                do_config_power_action (manager, GSD_POWER_ACTION_SUSPEND, power_action_noninteractive);
+                break;
         case HIBERNATE_KEY:
-        case POWER_KEY_NO_DIALOG:
-        case SLEEP_KEY_NO_DIALOG:
-        case SUSPEND_KEY_NO_DIALOG:
-        case HIBERNATE_KEY_NO_DIALOG:
+                do_config_power_action (manager, GSD_POWER_ACTION_HIBERNATE, power_action_noninteractive);
+                break;
         case SCREEN_BRIGHTNESS_UP_KEY:
         case SCREEN_BRIGHTNESS_DOWN_KEY:
         case KEYBOARD_BRIGHTNESS_UP_KEY:
         case KEYBOARD_BRIGHTNESS_DOWN_KEY:
         case KEYBOARD_BRIGHTNESS_TOGGLE_KEY:
+                do_brightness_action (manager, type);
+                break;
         case BATTERY_KEY:
+                do_battery_action (manager);
                 break;
         /* Note, no default so compiler catches missing keys */
         case CUSTOM_KEY:
@@ -2155,11 +2239,23 @@ do_action (GsdMediaKeysManager *manager,
 static void
 on_accelerator_activated (ShellKeyGrabber     *grabber,
                           guint                accel_id,
-                          guint                deviceid,
-                          guint                timestamp,
+                          GVariant            *parameters,
                           GsdMediaKeysManager *manager)
 {
+        GVariantDict dict;
         guint i;
+        guint deviceid;
+        guint timestamp;
+        guint mode;
+
+        g_variant_dict_init (&dict, parameters);
+
+        if (!g_variant_dict_lookup (&dict, "device-id", "u", &deviceid))
+              deviceid = 0;
+        if (!g_variant_dict_lookup (&dict, "timestamp", "u", &timestamp))
+              timestamp = GDK_CURRENT_TIME;
+        if (!g_variant_dict_lookup (&dict, "action-mode", "u", &mode))
+              mode = 0;
 
         for (i = 0; i < manager->priv->keys->len; i++) {
                 MediaKey *key;
@@ -2172,7 +2268,7 @@ on_accelerator_activated (ShellKeyGrabber     *grabber,
                 if (key->key_type == CUSTOM_KEY)
                         do_custom_action (manager, deviceid, key, timestamp);
                 else
-                        do_action (manager, deviceid, key->key_type, timestamp);
+                        do_action (manager, deviceid, mode, key->key_type, timestamp);
                 return;
         }
 }
@@ -2313,6 +2409,12 @@ start_media_keys_idle_cb (GsdMediaKeysManager *manager)
           g_hash_table_new_full (g_str_hash, g_str_equal,
                                  g_free, g_object_unref);
 
+#if 0
+        /* for the power plugin interface code */
+        manager->priv->power_settings = g_settings_new (SETTINGS_POWER_DIR);
+        manager->priv->chassis_type = gnome_settings_get_chassis_type ();
+#endif
+
         /* Logic from http://git.gnome.org/browse/gnome-shell/tree/js/ui/status/accessibility.js#n163 */
         manager->priv->interface_settings = g_settings_new (SETTINGS_INTERFACE_DIR);
         g_signal_connect (G_OBJECT (manager->priv->interface_settings), "changed::gtk-theme",
@@ -2413,6 +2515,14 @@ gsd_media_keys_manager_stop (GsdMediaKeysManager *manager)
 
         g_clear_object (&priv->logind_proxy);
         g_clear_object (&priv->settings);
+#if 0
+        g_clear_object (&priv->power_settings);
+        g_clear_object (&priv->power_proxy);
+        g_clear_object (&priv->power_screen_proxy);
+        g_clear_object (&priv->power_keyboard_proxy);
+        g_clear_object (&priv->composite_device);
+        g_clear_pointer (&priv->chassis_type, g_free);
+#endif
         g_clear_object (&priv->mpris_controller);
         g_clear_object (&priv->screencast_proxy);
 
@@ -2642,6 +2752,9 @@ on_bus_gotten (GObject             *source_object,
 {
         GDBusConnection *connection;
         GError *error = NULL;
+#if 0
+        UpClient *up_client;
+#endif
 
         if (manager->priv->bus_cancellable == NULL ||
             g_cancellable_is_cancelled (manager->priv->bus_cancellable)) {
@@ -2674,6 +2787,42 @@ on_bus_gotten (GObject             *source_object,
                           NULL,
                           (GAsyncReadyCallback) xrandr_ready_cb,
                           manager);
+
+#if 0
+        g_dbus_proxy_new (manager->priv->connection,
+                          G_DBUS_PROXY_FLAGS_NONE,
+                          NULL,
+                          GSD_DBUS_NAME ".Power",
+                          GSD_DBUS_PATH "/Power",
+                          GSD_DBUS_BASE_INTERFACE ".Power",
+                          NULL,
+                          (GAsyncReadyCallback) power_ready_cb,
+                          manager);
+
+        g_dbus_proxy_new (manager->priv->connection,
+                          G_DBUS_PROXY_FLAGS_NONE,
+                          NULL,
+                          GSD_DBUS_NAME ".Power",
+                          GSD_DBUS_PATH "/Power",
+                          GSD_DBUS_BASE_INTERFACE ".Power.Screen",
+                          NULL,
+                          (GAsyncReadyCallback) power_screen_ready_cb,
+                          manager);
+
+        g_dbus_proxy_new (manager->priv->connection,
+                          G_DBUS_PROXY_FLAGS_NONE,
+                          NULL,
+                          GSD_DBUS_NAME ".Power",
+                          GSD_DBUS_PATH "/Power",
+                          GSD_DBUS_BASE_INTERFACE ".Power.Keyboard",
+                          NULL,
+                          (GAsyncReadyCallback) power_keyboard_ready_cb,
+                          manager);
+
+        up_client = up_client_new ();
+        manager->priv->composite_device = up_client_get_display_device (up_client);
+        g_object_unref (up_client);
+#endif
 }
 
 static void
